--[[
MRPAS
Copyright (c) 2010 Dominik Marczuk
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of Dominik Marczuk may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY DOMINIK MARCZUK ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL DOMINIK MARCZUK BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
--]]

--[[ Love2d - Lua - Porting of Restrictive Precise Angle Shadowcasting by Mingo http://umbrarumregnum.110mb.com/download/mrpas --]]

--[[ Astar code by Deco Da Man http://www.facepunch.com/member.php?u=103897, original post http://wiki.garrysmod.com/?title=A-Star_Pathfinding_Module --]]

function distance(x1,y1,x2,y2)
  local dx,dy = x2-x1,y2-y1
  return math.sqrt(dx*dx+dy*dy)
end

function MyNeighbourIterator(node)
  
  neighbours={}
  function verifyneighbours(px,py)
    if px>=0 and px<=MAP_WIDTH then
      if py>=0 and py<=MAP_HEIGHT then
        if map[px][py].transparent == true then
				  --love.graphics.setColor(60,60,128,100)
				  --love.graphics.rectangle("line", px*10+2, py*10+2,6,6)
          table.insert(neighbours,map[px][py])
        end
      end
    end
  end
  
  --local dx = target_node.x-starting_node.x
  --local dy = target_node.y-starting_node.y
  
	verifyneighbours (node.x,node.y+1)
	verifyneighbours (node.x,node.y-1)
	verifyneighbours (node.x-1,node.y)
  verifyneighbours (node.x+1,node.y)
  verifyneighbours (node.x+1,node.y+1)
	verifyneighbours (node.x+1,node.y-1)
	verifyneighbours (node.x-1,node.y+1)
	verifyneighbours (node.x-1,node.y-1)
	
	return pairs(neighbours)
end

function MyWalkableCheck(current_node, neighbour_node, starting_node, target_node, heuristic, ...)
	return neighbour_node.transparent;
end

function MyHeuristic(node_a, node_b,starting_node, target_node)
	return distance(node_a.x,node_a.y,node_b.x,node_b.y)
end

function MyHeuristicCloserPath(node_a, node_b, starting_node, target_node)
	--local dx1 = node_b.x - target_node.x
	--local dy1 = node_b.y - target_node.y
	--local dx2 = starting_node.x - target_node.x
	--local dy2 = starting_node.y - target_node.y
	--return math.abs(dx1*dy2 - dx2*dy1)
	local dx1 = node_a.x - target_node.x
	local dy1 = node_a.y - target_node.y
	return math.abs(dx1*dx1 + dy1*dy1)
end

function MyConditional(current_node, neighbour_node, starting_node, target_node, hueristic, ...)
	--return neighbour_node.material == lava and 10 or 1 -- makes it consider alternatives and only cross the lava if there is no other path.
	-- Leo : magari se c' un mostro ritornare un 3 cos cerca di evitarlo...
	-- per adesso niente
	return 1
end

function booleano(pbool,pstringtrue,pstringfalse)
  if pbool == nil then
    return "NIL"
  elseif pbool then
    return pstringtrue
  else
    return pstringfalse
  end
end

function numerico(pnumero,stringa0,stringa1,stringa2)
  if pnumero == nil then
    return "NIL"
  elseif pnumero==0 then
    return stringa0
  elseif pnumero==1 then
    return stringa1
  elseif pnumero==2 then
    return stringa2
  end
end

function MyAstar(x1,y1,x2,y2)
  if map[x1][y1].transparent == false or map[x2][y2].transparent == false then
    return nil
  end
  if astar_closerpath==0 then
    local found,path,statuses, steps=astar.CalculatePath(map[x1][y1], map[x2][y2], MyNeighbourIterator, MyWalkableCheck, MyHeuristic, MyConditional)
    return path, steps
  elseif astar_closerpath==1 then
    local found,path,statuses, steps=astar.CalculatePath(map[x1][y1], map[x2][y2], MyNeighbourIterator, MyWalkableCheck, MyHeuristicCloserPath, MyConditional)
    return path, steps
  elseif astar_closerpath==2 then
    local steps=-2
    local path=calc(x1,y1,x2,y2)
    return path, steps
  end
end

function love.load()

 --profiler require "profiler"
 require "mrpas"
 require "los"
 --require "a-star"
 require "astar"
 require "Astar_tengine"

 MAP_WIDTH=80
 MAP_HEIGHT=60
 local j
 local k
 map = {}
 enemy = {}
 astars = {}
 astarssteps = {}
 
 for j=0, MAP_WIDTH, 1 do
		map[j] = {}
		for k=0, MAP_HEIGHT, 1 do
			map[j][k] = {x=j,y=k,type=0,fov=false,viewed=false,transparent=true,walkable=false}
		end
	end
	-- sample walls
	for k=4,50,1 do
	  map[k][10].type=1
	  map[k][10].transparent=false
	end
	for k=4,42,1 do
	  map[6][k+10].type=1
	  map[6][k+10].transparent=false
	end
	for k=40,48,1 do
	  map[k][40].type=1
	  map[k][40].transparent=false
	  map[k][48].type=1
	  map[k][48].transparent=false
	  map[40][k].type=1
	  map[40][k].transparent=false
	  map[48][k].type=1
	  map[48][k].transparent=false
	end
  map[40][44].type=0
	map[40][44].transparent=true
	--
	player_posx=40
	player_posy=25
	moved = false
	fovfps = false
	losfps = false
	astarfps = false
	enemymove = false
	astar_closerpath = 2
	--
	enemy[1]={type=1,x=20,y=20}
	enemy[2]={type=1,x=45,y=45}
	--
	bdrawing=0
	--
	computeFOV(map,player_posx,player_posy,0,true)
end

function love.keypressed(key)
	if key == " " then
		moved=true
	end
	if key == "1" then
		fovfps=false
	end
	if key == "2" then
		fovfps=true
	end
	if key == "3" then
		losfps=false
	end
	if key == "4" then
		losfps=true
	end
	if key == "5" then
		astarfps=false
	end
	if key == "6" then
		astarfps=true
	end
	if key == "7" then
		enemymove=true
	end
	if key == "c" then
	  if love.keyboard.isDown( 'lshift') then
		  astar_closerpath = 1
		elseif love.keyboard.isDown( 'lctrl') then
		  astar_closerpath = 2
    else
		  astar_closerpath = 0
		end
	end
	if key == "p" then
    if love.keyboard.isDown( 'lctrl') then
		  --profiler profiler.start("profiler.txt")
		end
	end
	if key == "o" then
    if love.keyboard.isDown( 'lctrl') then
		  --profiler profiler.stop()
		end
	end

	if key == "left" then
	  if player_posx > 0 then
	    player_posx = player_posx - 1
	    moved = true
	  end 
	end
	if key == "right" then
	  if player_posx < MAP_WIDTH-1 then
	    player_posx = player_posx + 1
	    moved = true
	  end 
	end
	if key == "up" then
	  if player_posy > 0 then
	    player_posy = player_posy - 1
	    moved = true
	  end 
	end
	if key == "down" then
	  if player_posy < MAP_HEIGHT-1 then
	    player_posy = player_posy + 1
	    moved = true
	  end 
	end
  --Debug (only with console)
  if key == "t" then 
    if love.keyboard.isDown( 'lctrl') then
      debug.debug()
    end
  end
end


function love.update()
  if moved or fovfps then
    computeFOV(map,player_posx,player_posy,0,true)
    moved=false
  end
  if enemymove then
  	for k,v in pairs(enemy) do
  	  --local astar=CalcMoves(map, v.x, v.y, player_posx, player_posy)
  	  --if astar ~= nil then
  	  --  if astar[2]~= nil then
  	  --    if map[astar[2].x][astar[2].y].transparent == true then
      --	    v.x = astar[2].x
  	  --      v.y = astar[2].y
  	  --    end
  	  --  end
  	  --end
  	  local astar, steps=MyAstar(v.x, v.y, player_posx, player_posy)
  	  if astar ~= nil then
  	    if astar[table.maxn(astar)-1]~= nil then
  	      if map[astar[table.maxn(astar)-1].x][astar[table.maxn(astar)-1].y].transparent == true then
      	    v.x = astar[table.maxn(astar)-1].x
  	        v.y = astar[table.maxn(astar)-1].y
  	      end
  	    end
  	  end
    end
    enemymove = false
  end
  if love.mouse.isDown('l') then
	  x, y = love.mouse.getPosition( )
	  nx = math.ceil((x-9)/10)
	  ny = math.ceil((y-9)/10)
	  if bdrawing == 0 then
	    if (map[nx][ny].type==0) then
	      bdrawing=1
	    else
	      bdrawing=2
	    end
	  end
	  if (bdrawing==1) then
	    map[nx][ny].type = 1
	    map[nx][ny].transparent = false
	  else
	    map[nx][ny].type = 0
	    map[nx][ny].transparent = true
	  end
	elseif love.mouse.isDown('r') then
	  bdrawing = 0
	  x, y = love.mouse.getPosition( )
	  nx = math.ceil((x-9)/10)
	  ny = math.ceil((y-9)/10)
	  local btrovato = false
	  local i
	  for i=#enemy ,1 , -1 do
	    local v=enemy[i]
		  if v.x == nx and v.y == ny then
		    btrovato = true
		    table.remove(enemy,i)
		  end
		end
		if btrovato == false then
		  table.insert(enemy,{type=1,x=nx,y=ny})
		end
	else
	  bdrawing = 0
	end
  computeFOV(map,player_posx,player_posy,0,true)
end

function love.draw()
  local j
  local k
  for j=0, MAP_WIDTH, 1 do
		for k=0, MAP_HEIGHT, 1 do
			if map[j][k].type == 0 then
			  if map[j][k].fov then
			    love.graphics.setColor(255,255,255,255)
			    love.graphics.rectangle("fill", j*10, k*10,10,10)
			  elseif map[j][k].viewed then
			    love.graphics.setColor(255,255,255,100)
			    love.graphics.rectangle("fill", j*10, k*10,10,10)
			  end
			elseif map[j][k].type == 1 then
			  if map[j][k].fov then
			    love.graphics.setColor(255,0,0,255)
			    love.graphics.rectangle("fill", j*10, k*10,10,10)
			  elseif map[j][k].viewed then
			    love.graphics.setColor(255,0,0,100)
			    love.graphics.rectangle("fill", j*10, k*10,10,10)
			  end
			end
		end
	end
	
	for k,v in pairs(enemy) do
		if map[v.x][v.y].fov == true then
			love.graphics.setColor(255,0,255,255)
			love.graphics.print("G", v.x*10, v.y*10+10)	
		else
		  -- still a little visible (debug and demo for visibility astar & los)
			love.graphics.setColor(255,255,255,128)
			love.graphics.print("G", v.x*10, v.y*10+10)	
		end
		if losfps then
		  if los(map,player_posx,player_posy,v.x,v.y) then
		    love.graphics.setColor(0,255,0,255)
		  else
		    love.graphics.setColor(255,0,0,255)
		  end 
		  love.graphics.line(player_posx*10+5,player_posy*10+5,v.x*10+5,v.y*10+5)
		end 
	end
  astars = {}
  astarssteps = {}
  if astarfps then
    for k,v in pairs(enemy) do
      --astars[k]=CalcMoves(map, v.x, v.y, player_posx, player_posy)
      astars[k],astarssteps[k]=MyAstar(v.x, v.y, player_posx, player_posy)
    end
  end 
	if astarfps then
	  for k,v in pairs(astars) do
	    if v ~= nil then
  	    --love.graphics.setColor(128,128,255,100)
	      --love.graphics.print(#v .." "..astarssteps[k], enemy[k].x*10+10,enemy[k].y*10+10)
	      for kk,vv in pairs(v) do
				  -- draw debug (remove it to use in projects)
				  love.graphics.setColor(128,128,255,100)
				  love.graphics.rectangle("line", vv.x*10+2, vv.y*10+2,6,6)
				  -- end draw debug
	      end
	    end
	  end
	end 
	
	love.graphics.setColor(255,0,0,255)
	love.graphics.print("@", player_posx*10, player_posy*10+10)
	--
	love.graphics.setColor(190,190,190,255)
	if fovfps then
    love.graphics.print(love.timer.getFPS().." fps, FOV every frame", 1, 12)
  else
    love.graphics.print(love.timer.getFPS().." fps, FOV after move", 1, 12)
  end
  love.graphics.print("mouse left click to toggle walls, mouse right click to toggle enemies", 1, 24)
  love.graphics.print("arrows to move @ character", 1, 36)
  love.graphics.print("1:calculate FOV after each move (default)    2:calculate FOV every update", 1, 48)
  love.graphics.print("3:don't calculate LOS enemies (default)      4:calculate LOS enemies", 1, 60)
  love.graphics.print("5:don't calculate a*star enemies (default)   6:calculate a*star enemies ".. numerico(astar_closerpath,"(normal)","(closer_path)","(tengine_closer_path)"), 1, 72)
  love.graphics.print("7:enemies make a move vs player", 1, 84)
end

